/*
 * kgr_entry_64_int3.S
 *
 * Entry code side of fix for CVE-2018-8897, bsc#1090368
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64) && !IS_ENABLED(CONFIG_XEN)

#include <linux/linkage.h>
#include <asm/dwarf2.h>
#include <asm/smap.h>
#include <asm/paravirt.h>
#include <asm/calling.h>
#include <asm/asm-offsets.h>
#include "kgr_patch_entry_code.h"


/* from arch/x86/kernel/entry_64.S */
	.macro KGR_EMPTY_FRAME offset=0
	CFI_STARTPROC simple
	CFI_SIGNAL_FRAME
	CFI_DEF_CFA rsp,\offset
	.endm

	.macro KGR_INTR_FRAME start=1 offset=0
	.if \start
	KGR_EMPTY_FRAME SS+8+\offset-RIP
	.else
	CFI_DEF_CFA_OFFSET SS+8+\offset-RIP
	.endif
	/*CFI_REL_OFFSET ss, SS+\offset-RIP*/
	CFI_REL_OFFSET rsp, RSP+\offset-RIP
	/*CFI_REL_OFFSET rflags, EFLAGS+\offset-RIP*/
	/*CFI_REL_OFFSET cs, CS+\offset-RIP*/
	CFI_REL_OFFSET rip, RIP+\offset-RIP
	.endm

	.macro KGR_XCPT_FRAME start=1 offset=0
	KGR_INTR_FRAME \start, RIP+\offset-ORIG_RAX
	.endm

	.macro KGR_PARTIAL_FRAME start=1 offset=0
	.if \start >= 0
	KGR_XCPT_FRAME \start, ORIG_RAX+\offset-ARGOFFSET
	.endif
	CFI_REL_OFFSET rdi, RDI+\offset-ARGOFFSET
	CFI_REL_OFFSET rsi, RSI+\offset-ARGOFFSET
	CFI_REL_OFFSET rdx, RDX+\offset-ARGOFFSET
	CFI_REL_OFFSET rcx, RCX+\offset-ARGOFFSET
	CFI_REL_OFFSET rax, RAX+\offset-ARGOFFSET
	CFI_REL_OFFSET r8, R8+\offset-ARGOFFSET
	CFI_REL_OFFSET r9, R9+\offset-ARGOFFSET
	CFI_REL_OFFSET r10, R10+\offset-ARGOFFSET
	CFI_REL_OFFSET r11, R11+\offset-ARGOFFSET
	.endm

	.macro KGR_DEFAULT_FRAME start=1 offset=0
	.if \start >= -1
	KGR_PARTIAL_FRAME \start, R11+\offset-R15
	.endif
	CFI_REL_OFFSET rbx, RBX+\offset
	CFI_REL_OFFSET rbp, RBP+\offset
	CFI_REL_OFFSET r12, R12+\offset
	CFI_REL_OFFSET r13, R13+\offset
	CFI_REL_OFFSET r14, R14+\offset
	CFI_REL_OFFSET r15, R15+\offset
	.endm


.macro kgr_zeroentry sym do_sym
ENTRY(kgr_repl_\sym)
	KGR_INTR_FRAME
	ASM_CLAC
	PARAVIRT_ADJUST_EXCEPTION_FRAME
	pushq_cfi $-1		/* ORIG_RAX: no syscall to restart */
	subq $ORIG_RAX-R15, %rsp
	CFI_ADJUST_CFA_OFFSET ORIG_RAX-R15
	KGR_CALL_PATCH kgr_error_entry
	KGR_DEFAULT_FRAME -1
	movq %rsp,%rdi		/* pt_regs pointer */
	xorl %esi,%esi		/* no error code */
	KGR_CALL_PATCH kgr_\do_sym
	KGR_JMP_PATCH kgr_error_exit		/* %ebx: no swapgs flag */
	CFI_ENDPROC
END(kgr_repl_\sym)
.endm



.code64
.text
.global __kgr_entry_text_begin
.align PAGE_SIZE
__kgr_entry_text_begin:

KGR_CALL_RELOCS_BEGIN


/*
 * Fix CVE-2018-8897
 *  -1 line, +1 line
 */
kgr_zeroentry int3 do_int3


KGR_CALL_RELOCS_END

.global __kgr_entry_text_end
.align PAGE_SIZE
__kgr_entry_text_end:

#endif /* IS_ENABLED(CONFIG_X86_64) && !IS_ENABLED(CONFIG_XEN) */
